@@ -1,6 +1,6 @@
 /*
 	Audio File Library
-	Copyright (C) 2011, Michael Pruett <michael@68k.org>
+	Copyright (C) 2011-2013, Michael Pruett <michael@68k.org>
 
 	This library is free software; you can redistribute it and/or
 	modify it under the terms of the GNU Library General Public
@@ -36,7 +36,8 @@
 const int _af_caf_compression_types[_AF_CAF_NUM_COMPTYPES] =
 {
 	AF_COMPRESSION_G711_ULAW,
-	AF_COMPRESSION_G711_ALAW
+	AF_COMPRESSION_G711_ALAW,
+	AF_COMPRESSION_IMA
 };
 
 enum
@@ -45,7 +46,7 @@ enum
 	kCAFLinearPCMFormatFlagIsLittleEndian = (1L << 1)
 };
 
-static _AFfilesetup caf_default_filesetup =
+static const _AFfilesetup cafDefaultFileSetup =
 {
 	_AF_VALID_FILESETUP,	// valid
 	AF_FILE_CAF,			// fileFormat
@@ -84,13 +85,13 @@ bool CAFFile::recognize(File *file)
 
 status CAFFile::readInit(AFfilesetup setup)
 {
-	fh->seek(8, File::SeekFromBeginning);
+	m_fh->seek(8, File::SeekFromBeginning);
 
-	trackCount = 1;
-	tracks = _af_track_new();
+	if (!allocateTrack())
+		return AF_FAIL;
 
-	off_t currentOffset = fh->tell();
-	off_t fileLength = fh->length();
+	off_t currentOffset = m_fh->tell();
+	off_t fileLength = m_fh->length();
 
 	while (currentOffset < fileLength)
 	{
@@ -106,8 +107,8 @@ status CAFFile::readInit(AFfilesetup setup)
 			chunkLength = fileLength - currentOffset;
 		else if (chunkLength < 0)
 			_af_error(AF_BAD_HEADER,
-				"invalid chunk length %lld for chunk type %s\n",
-				chunkLength, chunkType.name().c_str());
+				"invalid chunk length %jd for chunk type %s\n",
+				static_cast<intmax_t>(chunkLength), chunkType.name().c_str());
 
 		if (chunkType == "desc")
 		{
@@ -120,7 +121,7 @@ status CAFFile::readInit(AFfilesetup setup)
 				return AF_FAIL;
 		}
 
-		currentOffset = fh->seek(currentOffset + chunkLength,
+		currentOffset = m_fh->seek(currentOffset + chunkLength,
 			File::SeekFromBeginning);
 	}
 
@@ -129,13 +130,15 @@ status CAFFile::readInit(AFfilesetup setup)
 
 status CAFFile::writeInit(AFfilesetup setup)
 {
-	if (_af_filesetup_make_handle(setup, this) == AF_FAIL)
+	if (initFromSetup(setup) == AF_FAIL)
 		return AF_FAIL;
 
+	initCompressionParams();
+
 	Tag caff("caff");
 	if (!writeTag(&caff)) return AF_FAIL;
 	const uint8_t versionAndFlags[4] = { 0, 1, 0, 0 };
-	if (fh->write(versionAndFlags, 4) != 4) return AF_FAIL;
+	if (m_fh->write(versionAndFlags, 4) != 4) return AF_FAIL;
 
 	if (writeDescription() == AF_FAIL)
 		return AF_FAIL;
@@ -180,7 +183,8 @@ AFfilesetup CAFFile::completeSetup(AFfilesetup setup)
 
 	if (track->f.compressionType != AF_COMPRESSION_NONE &&
 		track->f.compressionType != AF_COMPRESSION_G711_ULAW &&
-		track->f.compressionType != AF_COMPRESSION_G711_ALAW)
+		track->f.compressionType != AF_COMPRESSION_G711_ALAW &&
+		track->f.compressionType != AF_COMPRESSION_IMA)
 	{
 		_af_error(AF_BAD_COMPTYPE,
 			"compression format %d not supported in CAF file",
@@ -188,7 +192,13 @@ AFfilesetup CAFFile::completeSetup(AFfilesetup setup)
 		return AF_NULL_FILESETUP;
 	}
 
-	return _af_filesetup_copy(setup, &caf_default_filesetup, true);
+	if (track->markersSet && track->markerCount)
+	{
+		_af_error(AF_BAD_NOT_IMPLEMENTED, "CAF does not yet support markers");
+		return AF_NULL_FILESETUP;
+	}
+
+	return _af_filesetup_copy(setup, &cafDefaultFileSetup, true);
 }
 
 status CAFFile::update()
@@ -203,21 +213,31 @@ status CAFFile::parseDescription(const Tag &, int64_t)
 	double sampleRate;
 	Tag formatID;
 	uint32_t formatFlags;
+	uint32_t bytesPerPacket;
+	uint32_t framesPerPacket;
 	uint32_t channelsPerFrame;
 	uint32_t bitsPerChannel;
 	if (!readDouble(&sampleRate) ||
 		!readTag(&formatID) ||
 		!readU32(&formatFlags) ||
-		!readU32(&m_bytesPerPacket) ||
-		!readU32(&m_framesPerPacket) ||
+		!readU32(&bytesPerPacket) ||
+		!readU32(&framesPerPacket) ||
 		!readU32(&channelsPerFrame) ||
 		!readU32(&bitsPerChannel))
 		return AF_FAIL;
 
+	if (!channelsPerFrame)
+	{
+		_af_error(AF_BAD_CHANNELS, "invalid file with 0 channels");
+		return AF_FAIL;
+	}
+
 	Track *track = getTrack();
 	track->f.channelCount = channelsPerFrame;
 	track->f.sampleWidth = bitsPerChannel;
 	track->f.sampleRate = sampleRate;
+	track->f.framesPerPacket = 1;
+
 	if (formatID == "lpcm")
 	{
 		track->f.compressionType = AF_COMPRESSION_NONE;
@@ -225,7 +245,7 @@ status CAFFile::parseDescription(const Tag &, int64_t)
 		{
 			if (bitsPerChannel != 32 && bitsPerChannel != 64)
 			{
-				_af_error(AF_BAD_WIDTH, "Invalid bits per sample %d for floating-point audio data", bitsPerChannel);
+				_af_error(AF_BAD_WIDTH, "invalid bits per sample %d for floating-point audio data", bitsPerChannel);
 				return AF_FAIL;
 			}
 			track->f.sampleFormat = bitsPerChannel == 32 ? AF_SAMPFMT_FLOAT :
@@ -238,7 +258,10 @@ status CAFFile::parseDescription(const Tag &, int64_t)
 		track->f.byteOrder = (formatFlags & kCAFLinearPCMFormatFlagIsLittleEndian) ?
 			AF_BYTEORDER_LITTLEENDIAN : AF_BYTEORDER_BIGENDIAN;
 
-		_af_set_sample_format(&track->f, track->f.sampleFormat, track->f.sampleWidth);
+		if (_af_set_sample_format(&track->f, track->f.sampleFormat, track->f.sampleWidth) == AF_FAIL)
+			return AF_FAIL;
+
+		track->f.computeBytesPerPacketPCM();
 		return AF_SUCCEED;
 	}
 	else if (formatID == "ulaw")
@@ -246,6 +269,7 @@ status CAFFile::parseDescription(const Tag &, int64_t)
 		track->f.compressionType = AF_COMPRESSION_G711_ULAW;
 		track->f.byteOrder = _AF_BYTEORDER_NATIVE;
 		_af_set_sample_format(&track->f, AF_SAMPFMT_TWOSCOMP, 16);
+		track->f.bytesPerPacket = channelsPerFrame;
 		return AF_SUCCEED;
 	}
 	else if (formatID == "alaw")
@@ -253,6 +277,15 @@ status CAFFile::parseDescription(const Tag &, int64_t)
 		track->f.compressionType = AF_COMPRESSION_G711_ALAW;
 		track->f.byteOrder = _AF_BYTEORDER_NATIVE;
 		_af_set_sample_format(&track->f, AF_SAMPFMT_TWOSCOMP, 16);
+		track->f.bytesPerPacket = channelsPerFrame;
+		return AF_SUCCEED;
+	}
+	else if (formatID == "ima4")
+	{
+		track->f.compressionType = AF_COMPRESSION_IMA;
+		track->f.byteOrder = _AF_BYTEORDER_NATIVE;
+		_af_set_sample_format(&track->f, AF_SAMPFMT_TWOSCOMP, 16);
+		initIMACompressionParams();
 		return AF_SUCCEED;
 	}
 	else
@@ -271,17 +304,13 @@ status CAFFile::parseData(const Tag &tag, int64_t length)
 
 	Track *track = getTrack();
 	if (length == -1)
-		track->data_size = fh->length() - fh->tell();
+		track->data_size = m_fh->length() - m_fh->tell();
 	else
 		track->data_size = length - 4;
-	track->fpos_first_frame = fh->tell();
-	int bytesPerFrame = track->f.bytesPerFrame(false);
-	if (track->f.compressionType == AF_COMPRESSION_G711_ULAW ||
-		track->f.compressionType == AF_COMPRESSION_G711_ALAW)
-	{
-		bytesPerFrame = track->f.channelCount;
-	}
-	track->totalfframes = track->data_size / bytesPerFrame;
+	track->fpos_first_frame = m_fh->tell();
+
+	track->computeTotalFileFrames();
+
 	return AF_SUCCEED;
 }
 
@@ -317,6 +346,14 @@ status CAFFile::writeDescription()
 		bytesPerPacket = channelsPerFrame;
 		bitsPerChannel = 8;
 	}
+	else if (track->f.compressionType == AF_COMPRESSION_IMA)
+	{
+		formatID = "ima4";
+		formatFlags = 0;
+		bytesPerPacket = track->f.bytesPerPacket;
+		framesPerPacket = track->f.framesPerPacket;
+		bitsPerChannel = 16;
+	}
 
 	if (!writeTag(&desc) ||
 		!writeS64(&chunkLength) ||
@@ -336,9 +373,9 @@ status CAFFile::writeData(bool update)
 	Track *track = getTrack();
 
 	if (m_dataOffset == -1)
-		m_dataOffset = fh->tell();
+		m_dataOffset = m_fh->tell();
 	else
-		fh->seek(m_dataOffset, File::SeekFromBeginning);
+		m_fh->seek(m_dataOffset, File::SeekFromBeginning);
 
 	Tag data("data");
 	int64_t dataLength = -1;
@@ -351,6 +388,29 @@ status CAFFile::writeData(bool update)
 		!writeU32(&editCount))
 		return AF_FAIL;
 	if (track->fpos_first_frame == 0)
-		track->fpos_first_frame = fh->tell();
+		track->fpos_first_frame = m_fh->tell();
 	return AF_SUCCEED;
 }
+
+void CAFFile::initCompressionParams()
+{
+	Track *track = getTrack();
+	if (track->f.compressionType == AF_COMPRESSION_IMA)
+		initIMACompressionParams();
+}
+
+void CAFFile::initIMACompressionParams()
+{
+	Track *track = getTrack();
+
+	track->f.bytesPerPacket = 34 * track->f.channelCount;
+	track->f.framesPerPacket = 64;
+
+	AUpvlist pv = AUpvnew(1);
+	AUpvsetparam(pv, 0, _AF_IMA_ADPCM_TYPE);
+	AUpvsetvaltype(pv, 0, AU_PVTYPE_LONG);
+	long l = _AF_IMA_ADPCM_TYPE_QT;
+	AUpvsetval(pv, 0, &l);
+
+	track->f.compressionParams = pv;
+}
