@@ -1,7 +1,8 @@
 /*
 	Audio File Library
-	Copyright (C) 1998-2000, Michael Pruett <michael@68k.org>
-	Copyright (C) 2000-2001, Silicon Graphics, Inc.
+	Copyright (C) 1998-2000, 2003-2004, 2010-2013, Michael Pruett <michael@68k.org>
+	Copyright (C) 2000-2002, Silicon Graphics, Inc.
+	Copyright (C) 2002-2003, Davy Durham
 
 	This library is free software; you can redistribute it and/or
 	modify it under the terms of the GNU Library General Public
@@ -32,7 +33,6 @@
 #include <assert.h>
 #include <math.h>
 #include <stdint.h>
-#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -96,7 +96,9 @@ enum
 const int _af_wave_compression_types[_AF_WAVE_NUM_COMPTYPES] =
 {
 	AF_COMPRESSION_G711_ULAW,
-	AF_COMPRESSION_G711_ALAW
+	AF_COMPRESSION_G711_ALAW,
+	AF_COMPRESSION_IMA,
+	AF_COMPRESSION_MS_ADPCM
 };
 
 const InstParamInfo _af_wave_inst_params[_AF_WAVE_NUM_INSTPARAMS] =
@@ -110,7 +112,7 @@ const InstParamInfo _af_wave_inst_params[_AF_WAVE_NUM_INSTPARAMS] =
 	{ AF_INST_NUMDBS_GAIN, AU_PVTYPE_LONG, "Gain in dB", {0} }
 };
 
-static _AFfilesetup wave_default_filesetup =
+static const _AFfilesetup waveDefaultFileSetup =
 {
 	_AF_VALID_FILESETUP,	/* valid */
 	AF_FILE_WAVE,		/* fileFormat */
@@ -150,11 +152,12 @@ WAVEFile::WAVEFile()
 {
 	setFormatByteOrder(AF_BYTEORDER_LITTLEENDIAN);
 
-	factOffset = 0;
-	miscellaneousStartOffset = 0;
-	totalMiscellaneousSize = 0;
-	markOffset = 0;
-	dataSizeOffset = 0;
+	m_factOffset = 0;
+	m_miscellaneousOffset = 0;
+	m_markOffset = 0;
+	m_dataSizeOffset = 0;
+
+	m_msadpcmNumCoefficients = 0;
 }
 
 status WAVEFile::parseFrameCount(const Tag &id, uint32_t size)
@@ -184,12 +187,19 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 	uint16_t blockAlign;
 	readU16(&blockAlign);
 
+	if (!channelCount)
+	{
+		_af_error(AF_BAD_CHANNELS, "invalid file with 0 channels");
+		return AF_FAIL;
+	}
+
 	track->f.channelCount = channelCount;
 	track->f.sampleRate = sampleRate;
 	track->f.byteOrder = AF_BYTEORDER_LITTLEENDIAN;
 
 	/* Default to uncompressed audio data. */
 	track->f.compressionType = AF_COMPRESSION_NONE;
+	track->f.framesPerPacket = 1;
 
 	switch (formatTag)
 	{
@@ -221,6 +231,7 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 			track->f.sampleFormat = AF_SAMPFMT_TWOSCOMP;
 			track->f.byteOrder = _AF_BYTEORDER_NATIVE;
 			track->f.compressionType = AF_COMPRESSION_G711_ULAW;
+			track->f.bytesPerPacket = track->f.channelCount;
 			break;
 
 		case WAVE_FORMAT_ALAW:
@@ -229,6 +240,7 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 			track->f.sampleFormat = AF_SAMPFMT_TWOSCOMP;
 			track->f.byteOrder = _AF_BYTEORDER_NATIVE;
 			track->f.compressionType = AF_COMPRESSION_G711_ALAW;
+			track->f.bytesPerPacket = track->f.channelCount;
 			break;
 
 		case WAVE_FORMAT_IEEE_FLOAT:
@@ -270,15 +282,12 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 			/* numCoefficients should be at least 7. */
 			assert(numCoefficients >= 7 && numCoefficients <= 255);
 
-			for (int i=0; i<numCoefficients; i++)
+			m_msadpcmNumCoefficients = numCoefficients;
+
+			for (int i=0; i<m_msadpcmNumCoefficients; i++)
 			{
-				int16_t a0, a1;
-
-				readS16(&a0);
-				readS16(&a1);
-
-				msadpcmCoefficients[i][0] = a0;
-				msadpcmCoefficients[i][1] = a1;
+				readS16(&m_msadpcmCoefficients[i][0]);
+				readS16(&m_msadpcmCoefficients[i][1]);
 			}
 
 			track->f.sampleWidth = 16;
@@ -286,30 +295,21 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 			track->f.compressionType = AF_COMPRESSION_MS_ADPCM;
 			track->f.byteOrder = _AF_BYTEORDER_NATIVE;
 
-			/* Create the parameter list. */
-			long l;
-			void *v;
-			AUpvlist pv = AUpvnew(4);
+			track->f.framesPerPacket = samplesPerBlock;
+			track->f.bytesPerPacket = blockAlign;
+
+			// Create the parameter list.
+			AUpvlist pv = AUpvnew(2);
 			AUpvsetparam(pv, 0, _AF_MS_ADPCM_NUM_COEFFICIENTS);
 			AUpvsetvaltype(pv, 0, AU_PVTYPE_LONG);
-			l = numCoefficients;
+			long l = m_msadpcmNumCoefficients;
 			AUpvsetval(pv, 0, &l);
 
 			AUpvsetparam(pv, 1, _AF_MS_ADPCM_COEFFICIENTS);
 			AUpvsetvaltype(pv, 1, AU_PVTYPE_PTR);
-			v = msadpcmCoefficients;
+			void *v = m_msadpcmCoefficients;
 			AUpvsetval(pv, 1, &v);
 
-			AUpvsetparam(pv, 2, _AF_FRAMES_PER_BLOCK);
-			AUpvsetvaltype(pv, 2, AU_PVTYPE_LONG);
-			l = samplesPerBlock;
-			AUpvsetval(pv, 2, &l);
-
-			AUpvsetparam(pv, 3, _AF_BLOCK_SIZE);
-			AUpvsetvaltype(pv, 3, AU_PVTYPE_LONG);
-			l = blockAlign;
-			AUpvsetval(pv, 3, &l);
-
 			track->f.compressionParams = pv;
 		}
 		break;
@@ -340,20 +340,10 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 			track->f.compressionType = AF_COMPRESSION_IMA;
 			track->f.byteOrder = _AF_BYTEORDER_NATIVE;
 
-			/* Create the parameter list. */
-			long l;
-			AUpvlist pv = AUpvnew(2);
-			AUpvsetparam(pv, 0, _AF_FRAMES_PER_BLOCK);
-			AUpvsetvaltype(pv, 0, AU_PVTYPE_LONG);
-			l = samplesPerBlock;
-			AUpvsetval(pv, 0, &l);
+			initIMACompressionParams();
 
-			AUpvsetparam(pv, 1, _AF_BLOCK_SIZE);
-			AUpvsetvaltype(pv, 1, AU_PVTYPE_LONG);
-			l = blockAlign;
-			AUpvsetval(pv, 1, &l);
-
-			track->f.compressionParams = pv;
+			track->f.framesPerPacket = samplesPerBlock;
+			track->f.bytesPerPacket = blockAlign;
 		}
 		break;
 
@@ -412,6 +402,7 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 				track->f.sampleWidth = 16;
 				track->f.sampleFormat = AF_SAMPFMT_TWOSCOMP;
 				track->f.byteOrder = _AF_BYTEORDER_NATIVE;
+				track->f.bytesPerPacket = channelCount;
 			}
 			else
 			{
@@ -445,6 +436,9 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 			break;
 	}
 
+	if (track->f.isUncompressed())
+		track->f.computeBytesPerPacketPCM();
+
 	_af_set_sample_format(&track->f, track->f.sampleFormat, track->f.sampleWidth);
 
 	return AF_SUCCEED;
@@ -452,11 +446,9 @@ status WAVEFile::parseFormat(const Tag &id, uint32_t size)
 
 status WAVEFile::parseData(const Tag &id, uint32_t size)
 {
-	assert(id == "data");
-
 	Track *track = getTrack();
 
-	track->fpos_first_frame = fh->tell();
+	track->fpos_first_frame = m_fh->tell();
 	track->data_size = size;
 
 	return AF_SUCCEED;
@@ -469,8 +461,8 @@ status WAVEFile::parsePlayList(const Tag &id, uint32_t size)
 
 	if (segmentCount == 0)
 	{
-		instrumentCount = 0;
-		instruments = NULL;
+		m_instrumentCount = 0;
+		m_instruments = NULL;
 		return AF_SUCCEED;
 	}
 
@@ -536,9 +528,9 @@ status WAVEFile::parseADTLSubChunk(const Tag &id, uint32_t size)
 {
 	Track *track = getTrack();
 
-	AFfileoffset endPos = fh->tell() + size;
+	AFfileoffset endPos = m_fh->tell() + size;
 
-	while (fh->tell() < endPos)
+	while (m_fh->tell() < endPos)
 	{
 		Tag chunkID;
 		uint32_t chunkSize;
@@ -553,7 +545,7 @@ status WAVEFile::parseADTLSubChunk(const Tag &id, uint32_t size)
 			char *p = (char *) _af_malloc(length);
 
 			readU32(&id);
-			fh->read(p, length);
+			m_fh->read(p, length);
 
 			Marker *marker = track->getMarker(id);
 
@@ -580,12 +572,12 @@ status WAVEFile::parseADTLSubChunk(const Tag &id, uint32_t size)
 				at the end of the chunk.
 			*/
 			if ((chunkSize % 2) != 0)
-				fh->seek(1, File::SeekFromCurrent);
+				m_fh->seek(1, File::SeekFromCurrent);
 		}
 		else
 		{
 			/* If chunkSize is odd, skip an extra byte. */
-			fh->seek(chunkSize + (chunkSize % 2), File::SeekFromCurrent);
+			m_fh->seek(chunkSize + (chunkSize % 2), File::SeekFromCurrent);
 		}
 	}
 	return AF_SUCCEED;
@@ -594,9 +586,9 @@ status WAVEFile::parseADTLSubChunk(const Tag &id, uint32_t size)
 /* Parse an INFO sub-chunk within a LIST chunk. */
 status WAVEFile::parseINFOSubChunk(const Tag &id, uint32_t size)
 {
-	AFfileoffset endPos = fh->tell() + size;
+	AFfileoffset endPos = m_fh->tell() + size;
 
-	while (fh->tell() < endPos)
+	while (m_fh->tell() < endPos)
 	{
 		int misctype = AF_MISC_UNRECOGNIZED;
 		Tag miscid;
@@ -622,25 +614,25 @@ status WAVEFile::parseINFOSubChunk(const Tag &id, uint32_t size)
 		{
 			char *string = (char *) _af_malloc(miscsize);
 
-			fh->read(string, miscsize);
+			m_fh->read(string, miscsize);
 
-			miscellaneousCount++;
-			miscellaneous = (Miscellaneous *) _af_realloc(miscellaneous, sizeof (Miscellaneous) * miscellaneousCount);
+			m_miscellaneousCount++;
+			m_miscellaneous = (Miscellaneous *) _af_realloc(m_miscellaneous, sizeof (Miscellaneous) * m_miscellaneousCount);
 
-			miscellaneous[miscellaneousCount-1].id = miscellaneousCount;
-			miscellaneous[miscellaneousCount-1].type = misctype;
-			miscellaneous[miscellaneousCount-1].size = miscsize;
-			miscellaneous[miscellaneousCount-1].position = 0;
-			miscellaneous[miscellaneousCount-1].buffer = string;
+			m_miscellaneous[m_miscellaneousCount-1].id = m_miscellaneousCount;
+			m_miscellaneous[m_miscellaneousCount-1].type = misctype;
+			m_miscellaneous[m_miscellaneousCount-1].size = miscsize;
+			m_miscellaneous[m_miscellaneousCount-1].position = 0;
+			m_miscellaneous[m_miscellaneousCount-1].buffer = string;
 		}
 		else
 		{
-			fh->seek(miscsize, File::SeekFromCurrent);
+			m_fh->seek(miscsize, File::SeekFromCurrent);
 		}
 
 		/* Make the current position an even number of bytes.  */
 		if (miscsize % 2 != 0)
-			fh->seek(1, File::SeekFromCurrent);
+			m_fh->seek(1, File::SeekFromCurrent);
 	}
 	return AF_SUCCEED;
 }
@@ -664,7 +656,7 @@ status WAVEFile::parseList(const Tag &id, uint32_t size)
 	else
 	{
 		/* Skip unhandled sub-chunks. */
-		fh->seek(size, File::SeekFromCurrent);
+		m_fh->seek(size, File::SeekFromCurrent);
 		return AF_SUCCEED;
 	}
 	return AF_SUCCEED;
@@ -711,22 +703,11 @@ status WAVEFile::readInit(AFfilesetup setup)
 
 	bool hasFormat = false;
 	bool hasData = false;
-	bool hasCue = false;
-	bool hasList = false;
-	bool hasPlayList = false;
 	bool hasFrameCount = false;
-	bool hasINST = false;
 
-	instruments = NULL;
-	instrumentCount = 0;
-	miscellaneous = NULL;
-	miscellaneousCount = 0;
+	Track *track = allocateTrack();
 
-	Track *track = _af_track_new();
-	tracks = track;
-	trackCount = 1;
-
-	fh->seek(0, File::SeekFromBeginning);
+	m_fh->seek(0, File::SeekFromBeginning);
 
 	readTag(&type);
 	readU32(&size);
@@ -735,10 +716,6 @@ status WAVEFile::readInit(AFfilesetup setup)
 	assert(type == "RIFF");
 	assert(formtype == "WAVE");
 
-#ifdef DEBUG
-	printf("size: %d\n", size);
-#endif
-
 	/* Include the offset of the form type. */
 	index += 4;
 
@@ -748,16 +725,9 @@ status WAVEFile::readInit(AFfilesetup setup)
 		uint32_t chunksize = 0;
 		status result;
 
-#ifdef DEBUG
-		printf("index: %d\n", index);
-#endif
 		readTag(&chunkid);
 		readU32(&chunksize);
 
-#ifdef DEBUG
-		printf("%s size: %d\n", chunkid.name().c_str(), chunksize);
-#endif
-
 		if (chunkid == "fmt ")
 		{
 			result = parseFormat(chunkid, chunksize);
@@ -796,28 +766,24 @@ status WAVEFile::readInit(AFfilesetup setup)
 		}
 		else if (chunkid == "cue ")
 		{
-			hasCue = true;
 			result = parseCues(chunkid, chunksize);
 			if (result == AF_FAIL)
 				return AF_FAIL;
 		}
 		else if (chunkid == "LIST" || chunkid == "list")
 		{
-			hasList = true;
 			result = parseList(chunkid, chunksize);
 			if (result == AF_FAIL)
 				return AF_FAIL;
 		}
 		else if (chunkid == "INST")
 		{
-			hasINST = true;
 			result = parseInstrument(chunkid, chunksize);
 			if (result == AF_FAIL)
 				return AF_FAIL;
 		}
 		else if (chunkid == "plst")
 		{
-			hasPlayList = true;
 			result = parsePlayList(chunkid, chunksize);
 			if (result == AF_FAIL)
 				return AF_FAIL;
@@ -829,7 +795,7 @@ status WAVEFile::readInit(AFfilesetup setup)
 		if ((index % 2) != 0)
 			index++;
 
-		fh->seek(index + 8, File::SeekFromBeginning);
+		m_fh->seek(index + 8, File::SeekFromBeginning);
 	}
 
 	/* The format chunk and the data chunk are required. */
@@ -839,26 +805,23 @@ status WAVEFile::readInit(AFfilesetup setup)
 	}
 
 	/*
-		At this point we know that the file has a format chunk
-		and a data chunk, so we can assume that track->f and
-		track->data_size have been initialized.
+		At this point we know that the file has a format chunk and a
+		data chunk, so we can assume that track->f and track->data_size
+		have been initialized.
 	*/
 	if (!hasFrameCount)
 	{
-		track->totalfframes = track->data_size /
-			(int) _af_format_frame_size(&track->f, false);
+		if (track->f.bytesPerPacket && track->f.framesPerPacket)
+		{
+			track->computeTotalFileFrames();
+		}
+		else
+		{
+			_af_error(AF_BAD_HEADER, "Frame count required but not found");
+			return AF_FAIL;
+		}
 	}
 
-	if (track->f.compressionType != AF_COMPRESSION_NONE &&
-		(track->f.compressionType == AF_COMPRESSION_G711_ULAW ||
-		track->f.compressionType == AF_COMPRESSION_G711_ALAW))
-	{
-		track->totalfframes = track->data_size / track->f.channelCount;
-	}
-
-	/*
-		A return value of AF_SUCCEED indicates successful parsing.
-	*/
 	return AF_SUCCEED;
 }
 
@@ -872,7 +835,14 @@ AFfilesetup WAVEFile::completeSetup(AFfilesetup setup)
 
 	TrackSetup *track = setup->getTrack();
 
-	if (track->sampleFormatSet)
+	if (track->f.isCompressed())
+	{
+		if (!track->sampleFormatSet)
+			_af_set_sample_format(&track->f, AF_SAMPFMT_TWOSCOMP, 16);
+		else
+			_af_set_sample_format(&track->f, track->f.sampleFormat, track->f.sampleWidth);
+	}
+	else if (track->sampleFormatSet)
 	{
 		switch (track->f.sampleFormat)
 		{
@@ -972,24 +942,25 @@ AFfilesetup WAVEFile::completeSetup(AFfilesetup setup)
 
 	if (track->f.compressionType != AF_COMPRESSION_NONE &&
 		track->f.compressionType != AF_COMPRESSION_G711_ULAW &&
-		track->f.compressionType != AF_COMPRESSION_G711_ALAW)
+		track->f.compressionType != AF_COMPRESSION_G711_ALAW &&
+		track->f.compressionType != AF_COMPRESSION_IMA &&
+		track->f.compressionType != AF_COMPRESSION_MS_ADPCM)
 	{
 		_af_error(AF_BAD_NOT_IMPLEMENTED, "compression format not supported in WAVE format");
 		return AF_NULL_FILESETUP;
 	}
 
-	if (track->byteOrderSet &&
+	if (track->f.isUncompressed() &&
+		track->byteOrderSet &&
 		track->f.byteOrder != AF_BYTEORDER_LITTLEENDIAN &&
-		track->f.compressionType == AF_COMPRESSION_NONE)
+		track->f.isByteOrderSignificant())
 	{
 		_af_error(AF_BAD_BYTEORDER, "WAVE format only supports little-endian data");
 		return AF_NULL_FILESETUP;
 	}
 
-	if (track->f.compressionType == AF_COMPRESSION_NONE)
+	if (track->f.isUncompressed())
 		track->f.byteOrder = AF_BYTEORDER_LITTLEENDIAN;
-	else
-		track->f.byteOrder = AF_BYTEORDER_BIGENDIAN;
 
 	if (track->aesDataSet)
 	{
@@ -1040,7 +1011,7 @@ AFfilesetup WAVEFile::completeSetup(AFfilesetup setup)
 	/*
 		Allocate an AFfilesetup and make all the unset fields correct.
 	*/
-	AFfilesetup	newsetup = _af_filesetup_copy(setup, &wave_default_filesetup, false);
+	AFfilesetup	newsetup = _af_filesetup_copy(setup, &waveDefaultFileSetup, false);
 
 	/* Make sure we do not copy loops if they are not specified in setup. */
 	if (setup->instrumentSet && setup->instrumentCount > 0 &&
@@ -1104,7 +1075,7 @@ status WAVEFile::writeFormat()
 
 	Track *track = getTrack();
 
-	fh->write("fmt ", 4);
+	m_fh->write("fmt ", 4);
 
 	switch (track->f.compressionType)
 	{
@@ -1147,6 +1118,20 @@ status WAVEFile::writeFormat()
 			bitsPerSample = 8;
 			break;
 
+		case AF_COMPRESSION_IMA:
+			chunkSize = 20;
+			formatTag = WAVE_FORMAT_DVI_ADPCM;
+			blockAlign = track->f.bytesPerPacket;
+			bitsPerSample = 4;
+			break;
+
+		case AF_COMPRESSION_MS_ADPCM:
+			chunkSize = 50;
+			formatTag = WAVE_FORMAT_ADPCM;
+			blockAlign = track->f.bytesPerPacket;
+			bitsPerSample = 4;
+			break;
+
 		default:
 			_af_error(AF_BAD_COMPTYPE, "bad compression type");
 			return AF_FAIL;
@@ -1163,9 +1148,12 @@ status WAVEFile::writeFormat()
 
 	averageBytesPerSecond =
 		track->f.sampleRate * _af_format_frame_size(&track->f, false);
+	if (track->f.compressionType == AF_COMPRESSION_IMA ||
+		track->f.compressionType == AF_COMPRESSION_MS_ADPCM)
+		averageBytesPerSecond = track->f.sampleRate * track->f.bytesPerPacket /
+			track->f.framesPerPacket;
 	writeU32(&averageBytesPerSecond);
 
-	blockAlign = _af_format_frame_size(&track->f, false);
 	writeU16(&blockAlign);
 
 	writeU16(&bitsPerSample);
@@ -1176,6 +1164,29 @@ status WAVEFile::writeFormat()
 		uint16_t zero = 0;
 		writeU16(&zero);
 	}
+	else if (track->f.compressionType == AF_COMPRESSION_IMA)
+	{
+		uint16_t extraByteCount = 2;
+		writeU16(&extraByteCount);
+		uint16_t samplesPerBlock = track->f.framesPerPacket;
+		writeU16(&samplesPerBlock);
+	}
+	else if (track->f.compressionType == AF_COMPRESSION_MS_ADPCM)
+	{
+		uint16_t extraByteCount = 2 + 2 + m_msadpcmNumCoefficients * 4;
+		writeU16(&extraByteCount);
+		uint16_t samplesPerBlock = track->f.framesPerPacket;
+		writeU16(&samplesPerBlock);
+
+		uint16_t numCoefficients = m_msadpcmNumCoefficients;
+		writeU16(&numCoefficients);
+
+		for (int i=0; i<m_msadpcmNumCoefficients; i++)
+		{
+			writeS16(&m_msadpcmCoefficients[i][0]);
+			writeS16(&m_msadpcmCoefficients[i][1]);
+		}
+	}
 
 	return AF_SUCCEED;
 }
@@ -1197,12 +1208,12 @@ status WAVEFile::writeFrameCount()
 		If the offset for the fact chunk hasn't been set yet,
 		set it to the file's current position.
 	*/
-	if (factOffset == 0)
-		factOffset = fh->tell();
+	if (m_factOffset == 0)
+		m_factOffset = m_fh->tell();
 	else
-		fh->seek(factOffset, File::SeekFromBeginning);
+		m_fh->seek(m_factOffset, File::SeekFromBeginning);
 
-	fh->write("fact", 4);
+	m_fh->write("fact", 4);
 	writeU32(&factSize);
 
 	totalFrameCount = track->totalfframes;
@@ -1215,13 +1226,13 @@ status WAVEFile::writeData()
 {
 	Track *track = getTrack();
 
-	fh->write("data", 4);
-	dataSizeOffset = fh->tell();
+	m_fh->write("data", 4);
+	m_dataSizeOffset = m_fh->tell();
 
 	uint32_t chunkSize = track->data_size;
 
 	writeU32(&chunkSize);
-	track->fpos_first_frame = fh->tell();
+	track->fpos_first_frame = m_fh->tell();
 
 	return AF_SUCCEED;
 }
@@ -1234,24 +1245,19 @@ status WAVEFile::update()
 	{
 		uint32_t dataLength, fileLength;
 
-		/* Update the frame count chunk if present. */
+		// Update the frame count chunk if present.
 		writeFrameCount();
 
-		/* Update the length of the data chunk. */
-		fh->seek(dataSizeOffset, File::SeekFromBeginning);
-
-		/*
-			We call _af_format_frame_size to calculate the
-			frame size of normal PCM data or compressed data.
-		*/
+		// Update the length of the data chunk.
+		m_fh->seek(m_dataSizeOffset, File::SeekFromBeginning);
 		dataLength = (uint32_t) track->data_size;
 		writeU32(&dataLength);
 
-		/* Update the length of the RIFF chunk. */
-		fileLength = (uint32_t) fh->length();
+		// Update the length of the RIFF chunk.
+		fileLength = (uint32_t) m_fh->length();
 		fileLength -= 8;
 
-		fh->seek(4, File::SeekFromBeginning);
+		m_fh->seek(4, File::SeekFromBeginning);
 		writeU32(&fileLength);
 	}
 
@@ -1262,7 +1268,7 @@ status WAVEFile::update()
 	*/
 	writeMiscellaneous();
 
-	/* Write the new positions; the size of the data will be unchanged. */
+	// Write the new positions; the size of the data will be unchanged.
 	writeCues();
 
 	return AF_SUCCEED;
@@ -1291,7 +1297,7 @@ static bool misc_type_to_wave (int misctype, Tag *miscid)
 
 status WAVEFile::writeMiscellaneous()
 {
-	if (miscellaneousCount != 0)
+	if (m_miscellaneousCount != 0)
 	{
 		uint32_t	miscellaneousBytes;
 		uint32_t 	chunkSize;
@@ -1300,31 +1306,29 @@ status WAVEFile::writeMiscellaneous()
 		miscellaneousBytes = 12;
 
 		/* Then calculate the size of the whole INFO chunk. */
-		for (int i=0; i<miscellaneousCount; i++)
+		for (int i=0; i<m_miscellaneousCount; i++)
 		{
 			Tag miscid;
 
 			// Skip miscellaneous data of an unsupported type.
-			if (!misc_type_to_wave(miscellaneous[i].type, &miscid))
+			if (!misc_type_to_wave(m_miscellaneous[i].type, &miscid))
 				continue;
 
 			// Account for miscellaneous type and size.
 			miscellaneousBytes += 8;
-			miscellaneousBytes += miscellaneous[i].size;
+			miscellaneousBytes += m_miscellaneous[i].size;
 
 			// Add a pad byte if necessary.
-			if (miscellaneous[i].size % 2 != 0)
+			if (m_miscellaneous[i].size % 2 != 0)
 				miscellaneousBytes++;
 
 			assert(miscellaneousBytes % 2 == 0);
 		}
 
-		if (miscellaneousStartOffset == 0)
-			miscellaneousStartOffset = fh->tell();
+		if (m_miscellaneousOffset == 0)
+			m_miscellaneousOffset = m_fh->tell();
 		else
-			fh->seek(miscellaneousStartOffset, File::SeekFromBeginning);
-
-		totalMiscellaneousSize = miscellaneousBytes;
+			m_fh->seek(m_miscellaneousOffset, File::SeekFromBeginning);
 
 		/*
 			Write the data.  On the first call to this
@@ -1336,46 +1340,46 @@ status WAVEFile::writeMiscellaneous()
 		*/
 
 		/* Write 'LIST'. */
-		fh->write("LIST", 4);
+		m_fh->write("LIST", 4);
 
 		/* Write the size of the following chunk. */
 		chunkSize = miscellaneousBytes-8;
 		writeU32(&chunkSize);
 
 		/* Write 'INFO'. */
-		fh->write("INFO", 4);
+		m_fh->write("INFO", 4);
 
 		/* Write each miscellaneous chunk. */
-		for (int i=0; i<miscellaneousCount; i++)
+		for (int i=0; i<m_miscellaneousCount; i++)
 		{
-			uint32_t miscsize = miscellaneous[i].size;
+			uint32_t miscsize = m_miscellaneous[i].size;
 			Tag miscid;
 
 			// Skip miscellaneous data of an unsupported type.
-			if (!misc_type_to_wave(miscellaneous[i].type, &miscid))
+			if (!misc_type_to_wave(m_miscellaneous[i].type, &miscid))
 				continue;
 
 			writeTag(&miscid);
 			writeU32(&miscsize);
-			if (miscellaneous[i].buffer != NULL)
+			if (m_miscellaneous[i].buffer != NULL)
 			{
 				uint8_t	zero = 0;
 
-				fh->write(miscellaneous[i].buffer, miscellaneous[i].size);
+				m_fh->write(m_miscellaneous[i].buffer, m_miscellaneous[i].size);
 
 				// Pad if necessary.
-				if ((miscellaneous[i].size%2) != 0)
+				if ((m_miscellaneous[i].size%2) != 0)
 					writeU8(&zero);
 			}
 			else
 			{
 				int	size;
-				size = miscellaneous[i].size;
+				size = m_miscellaneous[i].size;
 
 				// Pad if necessary.
 				if ((size % 2) != 0)
 					size++;
-				fh->seek(size, File::SeekFromCurrent);
+				m_fh->seek(size, File::SeekFromCurrent);
 			}
 		}
 	}
@@ -1385,140 +1389,138 @@ status WAVEFile::writeMiscellaneous()
 
 status WAVEFile::writeCues()
 {
-	int *markids, markCount;
-	uint32_t numCues, cueChunkSize, listChunkSize;
+	Track *track = getTrack();
 
-	markCount = afGetMarkIDs(this, AF_DEFAULT_TRACK, NULL);
-	if (markCount == 0)
+	if (!track->markerCount)
 		return AF_SUCCEED;
 
-	if (markOffset == 0)
-		markOffset = fh->tell();
+	if (m_markOffset == 0)
+		m_markOffset = m_fh->tell();
 	else
-		fh->seek(markOffset, File::SeekFromBeginning);
+		m_fh->seek(m_markOffset, File::SeekFromBeginning);
 
-	fh->write("cue ", 4);
+	Tag cue("cue ");
+	writeTag(&cue);
 
 	/*
 		The cue chunk consists of 4 bytes for the number of cue points
 		followed by 24 bytes for each cue point record.
 	*/
-	cueChunkSize = 4 + markCount * 24;
+	uint32_t cueChunkSize = 4 + track->markerCount * 24;
 	writeU32(&cueChunkSize);
-	numCues = markCount;
+	uint32_t numCues = track->markerCount;
 	writeU32(&numCues);
 
-	markids = (int *) _af_calloc(markCount, sizeof (int));
-	assert(markids != NULL);
-	afGetMarkIDs(this, AF_DEFAULT_TRACK, markids);
-
-	/* Write each marker to the file. */
-	for (int i=0; i < markCount; i++)
+	// Write each marker to the file.
+	for (int i=0; i<track->markerCount; i++)
 	{
-		uint32_t	identifier, position, chunkStart, blockStart;
-		uint32_t	sampleOffset;
-		AFframecount	markposition;
-
-		identifier = markids[i];
+		uint32_t identifier = track->markers[i].id;
 		writeU32(&identifier);
 
-		position = i;
+		uint32_t position = i;
 		writeU32(&position);
 
-		/* For now the RIFF id is always the first data chunk. */
-		fh->write("data", 4);
+		Tag data("data");
+		writeTag(&data);
 
 		/*
-			For an uncompressed WAVE file which contains
-			only one data chunk, chunkStart and blockStart
-			are zero.
+			For an uncompressed WAVE file which contains only one data chunk,
+			chunkStart and blockStart are zero.
 		*/
-		chunkStart = 0;
-		fh->write(&chunkStart, sizeof (uint32_t));
+		uint32_t chunkStart = 0;
+		writeU32(&chunkStart);
 
-		blockStart = 0;
-		fh->write(&blockStart, sizeof (uint32_t));
+		uint32_t blockStart = 0;
+		writeU32(&blockStart);
 
-		markposition = afGetMarkPosition(this, AF_DEFAULT_TRACK, markids[i]);
-
-		/* Sample offsets are stored in the WAVE file as frames. */
-		sampleOffset = markposition;
+		AFframecount markPosition = track->markers[i].position;
+		uint32_t sampleOffset = markPosition;
 		writeU32(&sampleOffset);
 	}
 
-	/*
-		Now write the cue names which is in a master list chunk
-		with a subchunk for each cue's name.
-	*/
-
-	listChunkSize = 4;
-	for (int i=0; i<markCount; i++)
+	// Now write the cue names and comments within a master list chunk.
+	uint32_t listChunkSize = 4;
+	for (int i=0; i<track->markerCount; i++)
 	{
-		const char *name;
-
-		name = afGetMarkName(this, AF_DEFAULT_TRACK, markids[i]);
+		const char *name = track->markers[i].name;
+		const char *comment = track->markers[i].comment;
 
 		/*
-			Each label chunk consists of 4 bytes for the
-			"labl" chunk ID, 4 bytes for the chunk data
-			size, 4 bytes for the cue point ID, and then
-			the length of the label as a Pascal-style string.
+			Each 'labl' or 'note' chunk consists of 4 bytes for the chunk ID,
+			4 bytes for the chunk data size, 4 bytes for the cue point ID,
+			and then the length of the label as a null-terminated string.
 
-			In all, this is 12 bytes plus the length of the
-			string, its size byte, and a trailing pad byte
-			if the length of the chunk is otherwise odd.
+			In all, this is 12 bytes plus the length of the string, its null
+			termination byte, and a trailing pad byte if the length of the
+			chunk is otherwise odd.
 		*/
-		listChunkSize += 12 + (strlen(name) + 1) +
-			((strlen(name) + 1) % 2);
+		listChunkSize += 12 + zStringLength(name);
+		listChunkSize += 12 + zStringLength(comment);
 	}
 
-	fh->write("LIST", 4);
+	Tag list("LIST");
+	writeTag(&list);
 	writeU32(&listChunkSize);
-	fh->write("adtl", 4);
+	Tag adtl("adtl");
+	writeTag(&adtl);
 
-	for (int i=0; i<markCount; i++)
+	for (int i=0; i<track->markerCount; i++)
 	{
-		const char	*name;
-		uint32_t	labelSize, cuePointID;
+		uint32_t cuePointID = track->markers[i].id;
 
-		name = afGetMarkName(this, AF_DEFAULT_TRACK, markids[i]);
-
-		/* Make labelSize even if it is not already. */
-		labelSize = 4+(strlen(name)+1) + ((strlen(name) + 1) % 2);
-		cuePointID = markids[i];
-
-		fh->write("labl", 4);
+		const char *name = track->markers[i].name;
+		uint32_t labelSize = 4 + zStringLength(name);
+		Tag lablTag("labl");
+		writeTag(&lablTag);
 		writeU32(&labelSize);
 		writeU32(&cuePointID);
-		fh->write(name, strlen(name) + 1);
-		/*
-			If the name plus the size byte comprises an odd
-			length, add another byte to make the string an
-			even length.
-		*/
-		if (((strlen(name) + 1) % 2) != 0)
-		{
-			uint8_t	zero=0;
-			writeU8(&zero);
-		}
-	}
+		writeZString(name);
 
-	free(markids);
+		const char *comment = track->markers[i].comment;
+		uint32_t noteSize = 4 + zStringLength(comment);
+		Tag noteTag("note");
+		writeTag(&noteTag);
+		writeU32(&noteSize);
+		writeU32(&cuePointID);
+		writeZString(comment);
+	}
 
 	return AF_SUCCEED;
 }
 
+bool WAVEFile::writeZString(const char *s)
+{
+	ssize_t lengthPlusNull = strlen(s) + 1;
+	if (m_fh->write(s, lengthPlusNull) != lengthPlusNull)
+		return false;
+	if (lengthPlusNull & 1)
+	{
+		uint8_t zero = 0;
+		if (!writeU8(&zero))
+			return false;
+	}
+	return true;
+}
+
+size_t WAVEFile::zStringLength(const char *s)
+{
+	size_t lengthPlusNull = strlen(s) + 1;
+	return lengthPlusNull + (lengthPlusNull & 1);
+}
+
 status WAVEFile::writeInit(AFfilesetup setup)
 {
-	uint32_t	zero = 0;
-
-	if (_af_filesetup_make_handle(setup, this) == AF_FAIL)
+	if (initFromSetup(setup) == AF_FAIL)
 		return AF_FAIL;
 
-	fh->seek(0, File::SeekFromBeginning);
-	fh->write("RIFF", 4);
-	fh->write(&zero, 4);
-	fh->write("WAVE", 4);
+	initCompressionParams();
+
+	uint32_t zero = 0;
+
+	m_fh->seek(0, File::SeekFromBeginning);
+	m_fh->write("RIFF", 4);
+	m_fh->write(&zero, 4);
+	m_fh->write("WAVE", 4);
 
 	writeMiscellaneous();
 	writeCues();
@@ -1531,10 +1533,69 @@ status WAVEFile::writeInit(AFfilesetup setup)
 
 bool WAVEFile::readUUID(UUID *u)
 {
-	return fh->read(u->data, 16) == 16;
+	return m_fh->read(u->data, 16) == 16;
 }
 
 bool WAVEFile::writeUUID(const UUID *u)
 {
-	return fh->write(u->data, 16) == 16;
+	return m_fh->write(u->data, 16) == 16;
+}
+
+void WAVEFile::initCompressionParams()
+{
+	Track *track = getTrack();
+	if (track->f.compressionType == AF_COMPRESSION_IMA)
+		initIMACompressionParams();
+	else if (track->f.compressionType == AF_COMPRESSION_MS_ADPCM)
+		initMSADPCMCompressionParams();
+}
+
+void WAVEFile::initIMACompressionParams()
+{
+	Track *track = getTrack();
+
+	track->f.framesPerPacket = 505;
+	track->f.bytesPerPacket = 256 * track->f.channelCount;
+
+	AUpvlist pv = AUpvnew(1);
+	AUpvsetparam(pv, 0, _AF_IMA_ADPCM_TYPE);
+	AUpvsetvaltype(pv, 0, AU_PVTYPE_LONG);
+	long l = _AF_IMA_ADPCM_TYPE_WAVE;
+	AUpvsetval(pv, 0, &l);
+
+	track->f.compressionParams = pv;
+}
+
+void WAVEFile::initMSADPCMCompressionParams()
+{
+	const int16_t coefficients[7][2] =
+	{
+		{ 256, 0 },
+		{ 512, -256 },
+		{ 0, 0 },
+		{ 192, 64 },
+		{ 240, 0 },
+		{ 460, -208 },
+		{ 392, -232 }
+	};
+	memcpy(m_msadpcmCoefficients, coefficients, sizeof (int16_t) * 7 * 2);
+	m_msadpcmNumCoefficients = 7;
+
+	Track *track = getTrack();
+
+	track->f.framesPerPacket = 500;
+	track->f.bytesPerPacket = 256 * track->f.channelCount;
+
+	AUpvlist pv = AUpvnew(2);
+	AUpvsetparam(pv, 0, _AF_MS_ADPCM_NUM_COEFFICIENTS);
+	AUpvsetvaltype(pv, 0, AU_PVTYPE_LONG);
+	long l = m_msadpcmNumCoefficients;
+	AUpvsetval(pv, 0, &l);
+
+	AUpvsetparam(pv, 1, _AF_MS_ADPCM_COEFFICIENTS);
+	AUpvsetvaltype(pv, 1, AU_PVTYPE_PTR);
+	void *v = m_msadpcmCoefficients;
+	AUpvsetval(pv, 1, &v);
+
+	track->f.compressionParams = pv;
 }
