@@ -32,9 +32,11 @@
 #include "audiofile.h"
 #include "error.h"
 
-typedef int status;
-#define AF_SUCCEED (0)
-#define AF_FAIL (-1)
+enum status
+{
+	AF_SUCCEED = 0,
+	AF_FAIL = -1
+};
 
 union AFPVu
 {
@@ -100,13 +102,14 @@ enum
 	_AF_WRITE_ACCESS = 2
 };
 
-/* The following are tokens for compression parameters in PV lists. */
+// The following are tokens for compression parameters in PV lists.
 enum
 {
-	_AF_FRAMES_PER_BLOCK = 700,		/* type: long */
-	_AF_BLOCK_SIZE = 701,			/* type: long */
 	_AF_MS_ADPCM_NUM_COEFFICIENTS = 800,	/* type: long */
-	_AF_MS_ADPCM_COEFFICIENTS = 801		/* type: array of int16_t[2] */
+	_AF_MS_ADPCM_COEFFICIENTS = 801,		/* type: array of int16_t[2] */
+	_AF_IMA_ADPCM_TYPE = 810,
+	_AF_IMA_ADPCM_TYPE_WAVE = 1,
+	_AF_IMA_ADPCM_TYPE_QT = 2
 };
 
 /* NeXT/Sun sampling rate */
