@@ -38,6 +38,7 @@
 #include "NeXT.h"
 #include "NIST.h"
 #include "Raw.h"
+#include "SampleVision.h"
 #include "VOC.h"
 #include "WAVE.h"
 
@@ -48,7 +49,7 @@
 #include "modules/IMA.h"
 #include "modules/MSADPCM.h"
 
-const _Unit _af_units[_AF_NUM_UNITS] =
+const Unit _af_units[_AF_NUM_UNITS] =
 {
 	{
 		AF_FILE_RAWDATA,
@@ -184,7 +185,17 @@ const _Unit _af_units[_AF_NUM_UNITS] =
 	{
 		AF_FILE_SAMPLEVISION,
 		"Sample Vision", "Sample Vision File Format", "smp",
-		false
+		true,
+		SampleVisionFile::completeSetup,
+		SampleVisionFile::recognize,
+		AF_SAMPFMT_TWOSCOMP, 16,
+		0,		// number of compression types
+		NULL,	// compression types
+		0,		// maximum marker count
+		0,		// maximum instrument count
+		0,		// maximum number of loops per instrument
+		0,		// number of instrument parameters
+		NULL	// instrument parameters
 	},
 	{
 		AF_FILE_VOC,
@@ -238,7 +249,7 @@ const _Unit _af_units[_AF_NUM_UNITS] =
 	}
 };
 
-const _CompressionUnit _af_compression[_AF_NUM_COMPRESSION] =
+const CompressionUnit _af_compression[_AF_NUM_COMPRESSION] =
 {
 	{
 		AF_COMPRESSION_NONE,
@@ -290,7 +301,7 @@ const _CompressionUnit _af_compression[_AF_NUM_COMPRESSION] =
 		true,	/* needsRebuffer */
 		false,	/* multiple_of */
 		_af_ima_adpcm_format_ok,
-		NULL, _af_ima_adpcm_init_decompress
+		_af_ima_adpcm_init_compress, _af_ima_adpcm_init_decompress
 	},
 	{
 		AF_COMPRESSION_MS_ADPCM,
@@ -303,6 +314,6 @@ const _CompressionUnit _af_compression[_AF_NUM_COMPRESSION] =
 		true,	/* needsRebuffer */
 		false,	/* multiple_of */
 		_af_ms_adpcm_format_ok,
-		NULL, _af_ms_adpcm_init_decompress
+		_af_ms_adpcm_init_compress, _af_ms_adpcm_init_decompress
 	}
 };
